--- old/usr/src/uts/common/fs/zfs/zio.c	2014-10-28 11:57:20.277340499 -0700
+++ new/usr/src/uts/common/fs/zfs/zio.c	2014-10-28 11:57:20.281453620 -0700
@@ -958,6 +958,12 @@
 {
 	blkptr_t *bp = zio->io_bp;
 
+	if (!BP_IS_EMBEDDED(bp) && BP_GET_PROP_RESERVATION(bp)) {
+		memset(zio->io_orig_data, 0, zio->io_orig_size);
+		zio->io_pipeline = ZIO_INTERLOCK_STAGES;
+		return (ZIO_PIPELINE_CONTINUE);
+	}
+
 	if (BP_GET_COMPRESS(bp) != ZIO_COMPRESS_OFF &&
 	    zio->io_child_type == ZIO_CHILD_LOGICAL &&
 	    !(zio->io_flags & ZIO_FLAG_RAW)) {
@@ -993,6 +999,8 @@
 	spa_t *spa = zio->io_spa;
 	zio_prop_t *zp = &zio->io_prop;
 	enum zio_compress compress = zp->zp_compress;
+	enum zio_checksum checksum = zp->zp_checksum;
+	uint8_t dedup = zp->zp_dedup;
 	blkptr_t *bp = zio->io_bp;
 	uint64_t lsize = zio->io_size;
 	uint64_t psize = lsize;
@@ -1011,6 +1019,12 @@
 
 	ASSERT(zio->io_child_type != ZIO_CHILD_DDT);
 
+	if (zp->zp_zero_write && !(zio->io_pipeline & ZIO_GANG_STAGES)) {
+		dedup = B_FALSE;
+		compress = ZIO_COMPRESS_OFF;
+		checksum = ZIO_CHECKSUM_OFF;
+	}
+
 	if (zio->io_bp_override) {
 		ASSERT(bp->blk_birth != zio->io_txg);
 		ASSERT(BP_GET_DEDUP(zio->io_bp_override) == 0);
@@ -1034,13 +1048,13 @@
 
 		ASSERT(!zp->zp_nopwrite);
 
-		if (BP_IS_HOLE(bp) || !zp->zp_dedup)
+		if (BP_IS_HOLE(bp) || !dedup)
 			return (ZIO_PIPELINE_CONTINUE);
 
-		ASSERT(zio_checksum_table[zp->zp_checksum].ci_dedup ||
+		ASSERT(zio_checksum_table[checksum].ci_dedup ||
 		    zp->zp_dedup_verify);
 
-		if (BP_GET_CHECKSUM(bp) == zp->zp_checksum) {
+		if (BP_GET_CHECKSUM(bp) == checksum) {
 			BP_SET_DEDUP(bp, 1);
 			zio->io_pipeline |= ZIO_STAGE_DDT_WRITE;
 			return (ZIO_PIPELINE_CONTINUE);
@@ -1153,6 +1167,17 @@
 		BP_SET_CHECKSUM(bp, zp->zp_checksum);
 		BP_SET_DEDUP(bp, zp->zp_dedup);
 		BP_SET_BYTEORDER(bp, ZFS_HOST_BYTEORDER);
+		if (zp->zp_zero_write  && !(zio->io_pipeline & ZIO_GANG_STAGES)) {
+			boolean_t need_allocate = B_FALSE;
+			if (zio->io_pipeline & ZIO_STAGE_DVA_ALLOCATE)
+				need_allocate = B_TRUE;
+			zio->io_pipeline = ZIO_INTERLOCK_STAGES;
+			if (need_allocate)
+				zio->io_pipeline |= ZIO_STAGE_DVA_ALLOCATE;
+			BP_SET_PROP_RESERVATION(bp, 1);
+		} else {
+			BP_SET_PROP_RESERVATION(bp, 0);
+		}
 		if (zp->zp_dedup) {
 			ASSERT(zio->io_child_type == ZIO_CHILD_LOGICAL);
 			ASSERT(!(zio->io_flags & ZIO_FLAG_IO_REWRITE));
@@ -1871,6 +1896,7 @@
 		zp.zp_copies = gio->io_prop.zp_copies;
 		zp.zp_dedup = B_FALSE;
 		zp.zp_dedup_verify = B_FALSE;
+		zp.zp_zero_write = B_FALSE;
 		zp.zp_nopwrite = B_FALSE;
 
 		zio_nowait(zio_write(zio, spa, txg, &gbh->zg_blkptr[g],
